#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


/* wmt_getspeed
 *
 * tool for reading current WM8505 clock speed, from the hardware registers
 * (does not require any kernel support)
 *
 * Usage : ./wmt_getspeed
 */
#define handle_error(msg)													\
  do { perror(msg); exit(EXIT_FAILURE); } while (0)

#define PMU_BASE  0xD8130000
#define OFFS_PLL_A 0x0200 / sizeof(unsigned int)
#define OFFS_PLL_B 0x0204 / sizeof(unsigned int)
#define OFFS_PLL_C 0x0208 / sizeof(unsigned int)
#define OFFS_CPU_DIV 0x0300 / sizeof(unsigned int)
#define OFFS_AHB_DIV 0x0304 / sizeof(unsigned int)
#define OFFS_RAM_DIV 0x0310 / sizeof(unsigned int)

#define PLL_PRE_DIV_BYPASS 1<<8  // Pre-divider halves source clock to PLL_A before use

#define PLL_MUL_MASK 0x1F
#define CPU_DIV_MASK 0x1F
#define AHB_DIV_MASK 0x07
#define RAM_DIV_MASK 0x1F

int reg_to_mult(unsigned int regval) {
  return 2 * (regval & PLL_MUL_MASK) / (regval & PLL_PRE_DIV_BYPASS ? 1 : 2);
}

int main(int argc, char *argv[])
{
  volatile unsigned int *addr;
  unsigned int pll_a, pll_b, pll_c, cpu_div, ahb_div, pll_mul, pll_pre_div, ram_div, pll_khz;
  unsigned int n,i;
  volatile double c;
  int fd;
  struct timeval start, now;
  
  fd = open("/dev/mem", O_RDWR|O_SYNC);
  if (fd == -1)
	 handle_error("open");
  
  addr = mmap(0, 0x1000, PROT_READ | PROT_WRITE, MAP_SHARED, fd, PMU_BASE);
  if(addr == (void *) -1)
	 handle_error("mmap");

  pll_a = addr[OFFS_PLL_A];
  pll_b = addr[OFFS_PLL_B];
  pll_c = addr[OFFS_PLL_C];
  cpu_div = addr[OFFS_CPU_DIV] & CPU_DIV_MASK;
  ahb_div = addr[OFFS_AHB_DIV] & AHB_DIV_MASK;
  ram_div = addr[OFFS_RAM_DIV] & RAM_DIV_MASK;
  
  printf("PLL_A mult 0x%02x (full reg 0x%02x) CPU div 0x%02x AHB div 0x%02x\n"
			"PLL B %02x (full reg 0x%02x)\n"
			"PLL C %02x (full reg 0x%02x) RAM div 0x%02x\n",
			reg_to_mult(pll_a), pll_a, cpu_div, ahb_div, 
			reg_to_mult(pll_b), pll_b,
			reg_to_mult(pll_c), pll_c, ram_div);

  pll_khz= 25 * 1000 * reg_to_mult(pll_a);
  printf("Estimated CPU speed %dkhz\n", pll_khz/cpu_div );
  printf("Estimated AHB speed %dkhz\n", pll_khz/cpu_div/ahb_div);
  printf("Estimated DDR RAM clock %dkhz\n", 25 * 1000 * reg_to_mult(pll_c)  / ram_div );

  printf("BogoCPUBench value ");

  // Extremely bogus CPU benchmark: how many times can we sum the series 1+1/2+1/3...1/100 in one second
  // (uses software FP so that should help slow it down a bit!)
  //
  // Bogus but should scale vaguely linearly with CPU clock speed
  //
  // Run at nice -n -19 to get a slightly more accurate value

  gettimeofday(&start, NULL);
  now = start;
  n = 0;
  while(now.tv_sec == start.tv_sec || now.tv_usec < start.tv_usec) {
	 c = 0.0;
	 for(i = 1; i <= 100; i++) {
		c += 1.0 / (float)i;
	 }
	 n++;
	 gettimeofday(&now, NULL);
  }
  
  printf("%d\n", n);

  exit(EXIT_SUCCESS);
} /* main */

